/***************************************************************************
                           mirror.cpp
                           ----------
    begin                : Sat Jul 31 2004
    copyright            : (C) 2004 by Dirk Ziegelmeier
    email                : dziegel@gmx.de
 ***************************************************************************/

/*
 * Mirror algorithm taken from DScaler.
 * Copyright (c) 2002 Rob Muller. All rights reserved.
 *
 * Ported by: Dirk Ziegelmeier
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 */

#include <kdebug.h>
#include <klocale.h>
#include <kconfig.h>

#include "mirror.h"
#include "kdetvimagefilter.h"
#include "kdetvimagefiltercontext.h"
#include "kdetvcpudetection.h"

// Qt 3.1.1 compatibility
#ifndef Q_INT64
#define Q_INT64 long long
#endif

// note that the pixels are stored as         Y0 C0a Y1 C0b Y2 C1a Y3 C1b (4 pixels)
// a swapped block of pixels looks like this: Y3 C1a Y2 C1b Y1 C0a Y0 C0b
#define SWAPPIXELS(pixelMM, temp1MM, temp2MM, CMask, MaskLumi02, MaskLumi13)                   \
    "movq    "pixelMM",    "temp1MM"\n\t"    /* make a copy */                                 \
    "psllq   $32,          "pixelMM"\n\t"    /* shift pixel 2 and 3 two pixels to the left */  \
    "psrlq   $32,          "temp1MM"\n\t"    /* shift pixel 0 and 1 two pixels to the right */ \
    "por     "temp1MM",    "pixelMM"\n\t"    /* combine the pixels again */                    \
    "movq    "pixelMM",    "temp2MM"\n\t"    /* make a copy */                                 \
    "movq    "pixelMM",    "temp1MM"\n\t"    /* make another copy */                           \
    "pand    "CMask",      "temp2MM"\n\t"    /* the chroma part of the pixels is ready now */  \
    "psllq   $16,          "pixelMM"\n\t"    /* shift 1 pixel to the left */                   \
    "psrlq   $16,          "temp1MM"\n\t"    /* shift 1 pixel to the right */                  \
    "pand    "MaskLumi02", "pixelMM"\n\t"    /* remove one pixel and all chroma*/              \
    "pand    "MaskLumi13", "temp1MM"\n\t"    /* remove one pixel and all chroma*/              \
    "por     "temp2MM",    "pixelMM"\n\t"    /* combine with the chroma */                     \
    "por     "temp1MM",    "pixelMM"\n\t"    /* combine the pixels */

class MirrorImageFilter : public KdetvImageFilter
{
public:
    MirrorImageFilter()
        : KdetvImageFilter("Mirror")
    {
    }

    virtual ~MirrorImageFilter()
    {
    }

    virtual KdetvImageFilterContext* operator<< (KdetvImageFilterContext* ctx);

    virtual KdetvImage::ImageFormat inputFormats()
    {
        return KdetvImage::FORMAT_YUV422;
    };
};

KdetvImageFilterContext* MirrorImageFilter::operator<< (KdetvImageFilterContext* ctx)
{
    // Return if we don't have MMX
    if(!(KdetvCpuDetection::capabilities() & KdetvCpuDetection::Cap_MMX)) {
        return ctx;
    }

    int            Cycles        = ctx->out->bytesPerLine() / 8 / 2; // half line in 8 byte chunks
    Q_INT64        qwMaskLumi02  = 0x00FF000000FF0000ull;
    Q_INT64        qwMaskLumi13  = 0x000000FF000000FFull;
    Q_INT64        qwCMask       = 0xFF00FF00FF00FF00ull;
    unsigned char* Pixels        = ctx->out->buffer();
    int            fieldHeight   = ctx->out->size().height();
    unsigned int   outputpitch   = ctx->out->bytesPerLine() + ctx->out->stride();

    #define _qwCMask      "%0"
    #define _qwMaskLumi02 "%1"
    #define _qwMaskLumi13 "%2"
    #define _Pixels       "%3"
    #define _Cycles       "%4"

    for (int y=0; y<fieldHeight; y++) {
        __asm__ __volatile__
            (
             "movq    "_qwCMask",      %%mm5\n\t"
             "movq    "_qwMaskLumi02", %%mm6\n\t"
             "movq    "_qwMaskLumi13", %%mm7\n\t"

             "movl    "_Pixels",       %%eax\n\t"     // eax point to the first pixel of the line
             "movl    "_Cycles",       %%edx\n\t"
             "shll    $4,              %%edx\n\t"
             "subl    $8,              %%edx\n\t"
             "addl    %%eax,           %%edx\n\t"     // edx now points to the last pixel of the line
             "movl    "_Cycles",       %%ecx\n\t"

             "1:\n\t"
             "movq    (%%eax),         %%mm0\n\t"     // get a pixelblock from the first half of the line
             "movq    (%%edx),         %%mm3\n\t"     // get the corresponding block from the second half

             SWAPPIXELS("%%mm0", "%%mm1", "%%mm2", "%%mm5", "%%mm6", "%%mm7") // swap the pixels from the first block

             "movq    %%mm0,           (%%edx)\n\t"   // and swap the block of pixels

             SWAPPIXELS("%%mm3", "%%mm1", "%%mm2", "%%mm5", "%%mm6", "%%mm7") // swap the pixels from the second block

             "movq    %%mm3,           (%%eax)\n\t"   // and swap the block of pixels

             "addl    $8,              %%eax\n\t"     // next pixel block
             "subl    $8,              %%edx\n\t"     // previous pixel block
             "decl    %%ecx\n\t"
             "jne     1b\n\t"

             : /* no outputs */

             : "m"(qwCMask),
               "m"(qwMaskLumi02),
               "m"(qwMaskLumi13),
               "m"(Pixels),
               "m"(Cycles)

             : "eax", "ecx", "edx",
               "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)",
               "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7",
               "memory", "cc"
             );

        Pixels += outputpitch;
    }

    __asm__ __volatile__ ("emms\n\t");

    return ctx;
}

// -----------------------------------------------------------------------

MirrorPlugin::MirrorPlugin(Kdetv *ktv, const QString& cfgkey, QObject *parent, const char* name)
    : KdetvFilterPlugin(ktv, cfgkey, parent, name)
{
    _filter = new MirrorImageFilter();
}

MirrorPlugin::~MirrorPlugin()
{
    delete _filter;
}

extern "C" {
    MirrorPlugin* create_mirror(Kdetv* ktv)
    {
        return new MirrorPlugin(ktv, "mirrorimage", 0, "Mirror image");
    }
}

#include "mirror.moc"
